/**
 * Copyright (c) 2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.admin.api;

import java.util.List;
import java.util.Properties;

import org.ow2.petals.admin.api.artifact.Logger;
import org.ow2.petals.admin.api.exception.ContainerAdministrationException;
import org.ow2.petals.admin.api.exception.InvalidLogLevelException;
import org.ow2.petals.admin.api.exception.LoggerNotFoundException;
import org.ow2.petals.admin.api.exception.NoConnectionException;
import org.ow2.petals.admin.topology.Domain;

/**
 * 
 * @author Nicolas Oddoux - EBM WebSourcing
 */
public interface ContainerAdministration {

    public void connect(String host, int port, String user, String password)
            throws ContainerAdministrationException;

    /**
     * @return <code>true</code> if connected to a Petals ESB container,
     *         <code>false</code> otherwise.
     * @throws ContainerAdministrationException
     *             An error occurs checking the connection.
     */
    public boolean isConnected() throws ContainerAdministrationException;

    /**
     * @throws NoConnectionException
     *             There is no current JMX connection to the remote petals ESB.
     * @throws ContainerAdministrationException
     *             An error occurs releasing the connection.
     */
    public void disconnect() throws NoConnectionException, ContainerAdministrationException;

    public String getSystemInfo() throws ContainerAdministrationException;

    public void stopContainer() throws ContainerAdministrationException;

    public void shutdownContainer() throws ContainerAdministrationException;

    /**
     * <p>
     * Get the current topology, including JBI artifacts deployed, filtered on:
     * <ul>
     * <li>containers, according to the regexp <code>containerFilterRegexp</code>.</li>
     * </ul>
     * </p>
     * 
     * @param containerFilterRegexp
     *            A container name regexp to filter the containers to return
     * @param securityPassPhrase
     *            The security pass-phrase to return sensible values
     * @param retrieveJbiArtifacts
     *            If <code>true</code>, the JBI artifacts deployed on each Petals container of the topology will be
     *            retrieved.
     * @return The current topology
     * @throws ContainerAdministrationException
     *             An error occurs
     */
    public Domain getTopology(final String containerFilterRegexp, final String securityPassPhrase,
            final boolean retrieveJbiArtifacts) throws ContainerAdministrationException;

    public Properties getServerProperties() throws ContainerAdministrationException;

    /**
     * Change the level of a logger.
     * 
     * @param loggerName
     *            The logger for which the level will be updated.
     * @param loggerLevel
     *            The level to set to the logger
     * @throws InvalidLogLevelException
     *             The level is invalid (does not exist).
     * @throws LoggerNotFoundException
     *             The logger does not exist.
     * @throws ContainerAdministrationException
     *             An error occurs updating the logger level.
     */
    public void changeLoggerLevel(String loggerName, String loggerLevel)
            throws InvalidLogLevelException, LoggerNotFoundException,
            ContainerAdministrationException;

    /**
     * Get loggers available on the container.
     * 
     * @return The list of loggers available on the container.
     * @throws ContainerAdministrationException
     *             An error occurs getting loggers.
     */
    public List<Logger> getLoggers() throws ContainerAdministrationException;

    /**
     * Attach the current container to another domain
     * 
     * @param targetDomainName
     *            The target domain name, the one that will be joined
     * @param targetContainerName
     *            The name of the container of the target domain, used as entry point in the target domain
     * @param targetContainerHost
     *            The host name of the container of the target domain
     * @param targetContainerPort
     *            The JMX port of the container of the target domain
     * @param targetContainerUser
     *            The JMX username of the container of the target domain
     * @param targetContainerPwd
     *            The JMX password of the container of the target domain
     * @param securityPassPhrase
     *            The security pass-phrase of the target container to access sensible information.
     * @throws ContainerAdministrationException
     *             An error occurs moving the current container
     */
    public void attachContainer(final String targetDomainName, final String targetContainerName,
            final String targetContainerHost, final int targetContainerPort, final String targetContainerUser,
            final String targetContainerPwd, final String securityPassPhrase) throws ContainerAdministrationException;

    /**
     * Detach the current container to another domain
     */
    public void detachContainer() throws ContainerAdministrationException;
}
