/**
 * Copyright (c) 2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.admin.api.artifact;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.ArrayList;
import java.util.Properties;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.apache.commons.io.IOUtils;
import org.ow2.petals.admin.api.artifact.Component.ComponentType;
import org.ow2.petals.admin.api.artifact.exception.ArtifactNameMissingInJBIDescriptorException;
import org.ow2.petals.admin.api.artifact.exception.ArtifactSlVersionMissingInJBIDescriptorException;
import org.ow2.petals.admin.api.artifact.exception.HttpCodeNotOkException;
import org.ow2.petals.admin.api.artifact.exception.HttpRemoteArtifactException;
import org.ow2.petals.admin.api.artifact.exception.InvalidContentTypeException;
import org.ow2.petals.admin.api.artifact.exception.InvalidZipArchiveException;
import org.ow2.petals.admin.api.artifact.exception.JBIDescriptorMissingException;
import org.ow2.petals.admin.api.artifact.exception.MissingContentTypeException;
import org.ow2.petals.admin.api.artifact.exception.SeveralServicesException;
import org.ow2.petals.admin.api.artifact.exception.TargetComponentMissingException;
import org.ow2.petals.jbi.descriptor.original.JBIDescriptorBuilder;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;

/**
 * 
 * @author Nicolas Oddoux - EBM WebSourcing
 */
public final class ArtifactUtils {

    /**
     * Property key to set {@link ArtifactUtils} implementation name
     */
    public static final String FACTORY_PROPERTY_KEY = "org.ow2.petals.admin.api.artifact.factory";

    /**
     * XPath expression to determine if a JBI configuration is tied to a binding
     * component
     */
    private static final String XPATH_JBI_BC_TYPE = "/*[local-name()='jbi']/*[local-name()='component'][@type='binding-component']";

    /**
     * XPath expression to determine if a JBI configuration is tied to a shared
     * library
     */
    private static final String XPATH_JBI_SL_TYPE = "/*[local-name()='jbi']/*[local-name()='shared-library']";

    /**
     * XPath expression to determine if a JBI configuration is tied to a service
     * assembly
     */
    private static final String XPATH_JBI_SA_TYPE = "/*[local-name()='jbi']/*[local-name()='service-assembly']";

    /**
     * XPath expression to determine if a JBI configuration is tied to a service unit
     */
    private static final String XPATH_JBI_SU_TYPE = "/*[local-name()='jbi']/*[local-name()='services']";

    /**
     * XPath expression to determine the number of services into a service unit
     */
    private static final String XPATH_JBI_SU_SERVICES_NB = "count(/*[local-name()='jbi']/*[local-name()='services']/*[local-name()='provides' or local-name()='consumes'])";

    /**
     * XPath expression to determine the target component name of a deployable service unit
     */
    private static final String XPATH_JBI_SU_TARGET_COMPONENT = "/*[local-name()='jbi']/*[local-name()='services']/*[local-name()='component-name']";

    /**
     * XPath expression to determine if a JBI configuration is tied to a service engine
     */
    private static final String XPATH_JBI_SE_TYPE = "/*[local-name()='jbi']/*[local-name()='component'][@type='service-engine']";

    /**
     * XPath expression to get all service unit name declared in a
     * ServiceAssembly descriptor.
     */
    private static final String XPATH_JBI_SU_NAME_AND_TARGETCOMPONENT_LIST = "//*[local-name()='service-unit']/*[local-name()='identification']/*[local-name()='name'] | //*[local-name()='service-unit']/*[local-name()='target']/*[local-name()='component-name']";

    /**
     * XPath expression to access the name of a JBI artifact
     */
    private static final String XPATH_JBI_NAME = "//*[local-name()='identification']/*[local-name()='name']";

    /**
     * XPath expression to access the version of a JBI shared library artifact
     */
    private static final String XPATH_JBI_SL_VERSION = "//*[local-name()='shared-library']/@version";

    private ArtifactUtils() {
        // NOP: It's an utility class
    }

    /**
     * Create artifact from the specified URL
     * 
     * @param artifactUrl
     * @return instance of artifact
     * @throws ArtifactException
     */
    public static final Artifact createArtifact(final URL artifactUrl) throws ArtifactException {
        assert artifactUrl != null;
        return createArtifact(artifactUrl, new Properties());
    }

    /**
     * Create artifact from the specified URL with the specified properties
     * files (Only useful for {@link Component})
     * 
     * @param artifactUrl
     * @param configurationFile
     * @return The artifact associated to the URL
     * @throws ArtifactException
     *             An error occurs reading the JBI archive or the properties
     *             file
     */
    public static final Artifact createArtifact(URL artifactUrl, URL configurationFile)
            throws ArtifactException {
        assert artifactUrl != null;
        assert configurationFile != null;

        try {
            final Properties configurationProperties = new Properties();
            final File tmpConfigurationFile = downloadPropertiesFileToTempZipFile(configurationFile);
            try (final Reader fileReader = new FileReader(tmpConfigurationFile)) {
                configurationProperties.load(fileReader);
            }
            return createArtifact(artifactUrl, configurationProperties);
        } catch (final IOException e) {
            throw new ArtifactException(e);
        }
    }

    /**
     * <p>
     * Create artifact from the specified URL with the specified properties files (Only useful for {@link Component})
     * </p>
     * <p>
     * Configuration parameters are given through {@link Properties}.
     * </p>
     * 
     * @param artifactUrl
     * @param configurationProperties
     *            Configuration parameters given as {@link Properties}. Not <code>null</code>.
     * @return
     * @throws ArtifactException
     */
    public static final Artifact createArtifact(final URL artifactUrl, final Properties configurationProperties)
            throws ArtifactException {
        assert configurationProperties != null;

        return createArtifact(artifactUrl, null, configurationProperties);
    }

    /**
     * <p>
     * Create artifact from the specified URL with the specified properties files (Only useful for {@link Component}).
     * </p>
     * <p>
     * Configuration parameters are given through a callback.
     * </p>
     * 
     * @param artifactUrl
     * @param configurationPropertiesCallback
     *            A call back to compute configuration parameters from the one already configured in the component. Not
     *            <code>null</code>.
     * @return
     * @throws ArtifactException
     */
    public static final Artifact createArtifact(final URL artifactUrl,
            final ConfigurationPropertiesCallback configurationPropertiesCallback)
            throws ArtifactException {
        assert configurationPropertiesCallback != null;

        return createArtifact(artifactUrl, configurationPropertiesCallback, null);
    }

    /**
     * <p>
     * Create artifact from the specified URL with the specified properties files (Only useful for {@link Component}).
     * </p>
     * <p>
     * Configuration parameters are given by <code>configurationPropertiesCallback</code> or
     * <code>configurationProperties</code>, not both in the same time.
     * </p>
     * 
     * @param artifactUrl
     * @param configurationPropertiesCallback
     *            A call back to compute configuration parameters from the one already configured in the component. Can
     *            be <code>null</code> if <code>configurationProperties</code> is not <code>null</code>.
     * @param configurationProperties
     *            Configuration parameters given as {@link Properties}. Can be <code>null</code> if
     *            <code>configurationPropertiesCallback</code> is not <code>null</code>.
     * @return
     * @throws ArtifactException
     */
    private static final Artifact createArtifact(final URL artifactUrl,
            final ConfigurationPropertiesCallback configurationPropertiesCallback,
            final Properties configurationProperties) throws ArtifactException {
        assert artifactUrl != null;
        assert configurationPropertiesCallback == null ^ configurationProperties == null;

        try {
            final File tmpJbiFile = downloadArtifactToTempZipFile(artifactUrl);
            Node jbiConfig;
            try {
                jbiConfig = extractJbiXml(tmpJbiFile);
            } catch (final JBIDescriptorMissingException e) {
                // Use the right URL, not the temporary one
                throw new JBIDescriptorMissingException(artifactUrl);
            }
            tmpJbiFile.delete();

            final XPath xpath = XPathFactory.newInstance().newXPath();
            if (((Boolean) xpath.evaluate(XPATH_JBI_BC_TYPE, jbiConfig, XPathConstants.BOOLEAN))
                    .booleanValue()) {
                if (configurationPropertiesCallback != null) {
                    return new Component(getName(jbiConfig, artifactUrl), ComponentType.BC,
                            configurationPropertiesCallback);
                } else {
                    return new Component(getName(jbiConfig, artifactUrl), ComponentType.BC, configurationProperties);
                }
            } else if (((Boolean) xpath.evaluate(XPATH_JBI_SE_TYPE, jbiConfig,
                    XPathConstants.BOOLEAN)).booleanValue()) {
                if (configurationPropertiesCallback != null) {
                    return new Component(getName(jbiConfig, artifactUrl), ComponentType.SE,
                            configurationPropertiesCallback);
                } else {
                    return new Component(getName(jbiConfig, artifactUrl), ComponentType.SE, configurationProperties);
                }
            } else if (((Boolean) xpath.evaluate(XPATH_JBI_SL_TYPE, jbiConfig,
                    XPathConstants.BOOLEAN)).booleanValue()) {
                return new SharedLibrary(getName(jbiConfig, artifactUrl), getSlVersion(jbiConfig, artifactUrl));
            } else if (((Boolean) xpath.evaluate(XPATH_JBI_SA_TYPE, jbiConfig,
                    XPathConstants.BOOLEAN)).booleanValue()) {
                final NodeList lst = (NodeList) xpath.evaluate(
                        XPATH_JBI_SU_NAME_AND_TARGETCOMPONENT_LIST, jbiConfig,
                        XPathConstants.NODESET);
                final ArrayList<ServiceUnit> sus = new ArrayList<ServiceUnit>();
                for (int i = 0; i < lst.getLength(); i = i + 2) {
                    final Node suName = lst.item(i);
                    final Node targetComponentName = lst.item(i + 1);
                    sus.add(new ServiceUnit(suName.getTextContent(), targetComponentName.getTextContent()));
                }
                return new ServiceAssembly(getName(jbiConfig, artifactUrl), sus);
            } else if (((Boolean) xpath.evaluate(XPATH_JBI_SU_TYPE, jbiConfig, XPathConstants.BOOLEAN)).booleanValue()) {

                if (((Double) xpath.evaluate(XPATH_JBI_SU_SERVICES_NB, jbiConfig, XPathConstants.NUMBER)).intValue() > 1) {
                    throw new SeveralServicesException(artifactUrl);
                } else {
                    final String targetComponent = (String) xpath.evaluate(XPATH_JBI_SU_TARGET_COMPONENT, jbiConfig,
                            XPathConstants.STRING);
                    if (targetComponent == null || targetComponent.isEmpty()) {
                        throw new TargetComponentMissingException(artifactUrl);
                    } else {
                        return new ServiceUnit(getName(jbiConfig, artifactUrl), targetComponent);
                    }
                }

            } else {
                final String errorMessage = String.format("%s: not a valid JBI archive",
                        artifactUrl.toString());
                throw new InvalidZipArchiveException(errorMessage, artifactUrl);
            }
        } catch (final IOException e) {
            throw new ArtifactException(e);
        } catch (final SAXException e) {
            throw new ArtifactException(e);
        } catch (final ParserConfigurationException e) {
            throw new ArtifactException(e);
        } catch (final XPathExpressionException e) {
            throw new ArtifactException(e);
        } catch (final DOMException e) {
            throw new ArtifactException(e);
        }
    }

    /**
     * Get the JBI descriptor of the provided JBI archive.
     * 
     * @param archive
     *            The JBI archive
     * @return The JBI descriptor of the provided JBI archive, as {@link Node}.
     * 
     * @throws JBIDescriptorMissingException
     *             The ZIP archive is invalid because the JBI descriptor is missing.
     * @throws IOException
     *             An error occurs reading the JBI archive or the JBI descriptor.
     * @throws ParserConfigurationException
     *             A error occurs creating the DOM document associated to the JBI descriptor.
     * @throws SAXException
     *             An error occurs parsing the JBI descriptor.
     */
    private static final Node extractJbiXml(final File archive) throws ParserConfigurationException, IOException,
            SAXException, JBIDescriptorMissingException {
        final ZipFile zipfile = new ZipFile(archive);
        try {
            final ZipEntry jbixml = zipfile.getEntry(JBIDescriptorBuilder.JBI_DESCRIPTOR_RESOURCE);
            if (jbixml != null) {
                final DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
                final DocumentBuilder db = dbf.newDocumentBuilder();
                final Document doc = db.parse(zipfile.getInputStream(jbixml));
                doc.getDocumentElement().normalize();
                return doc;
            } else {
                // The JBI descriptor is missing into the JBI archive
                throw new JBIDescriptorMissingException(archive.toURI().toURL());
            }
        } finally {
            zipfile.close();
        }
    }

    private static final File downloadPropertiesFileToTempZipFile(final URL propertiesUrl)
            throws HttpRemoteArtifactException, IOException {

        final InputStream propertiesStream;
        if ("http".equalsIgnoreCase(propertiesUrl.getProtocol())) {
            final HttpURLConnection httpUrlConnection = (HttpURLConnection) propertiesUrl.openConnection();
            final int responseCode = httpUrlConnection.getResponseCode();
            if (responseCode != HttpURLConnection.HTTP_OK) {
                throw new HttpCodeNotOkException(propertiesUrl, responseCode);
            } else {
                // We check that the resource is a binary file
                final String contentType = httpUrlConnection.getContentType();
                if (contentType != null) {
                    if (!contentType.startsWith(MimeTypeConstants.TEXT_PLAIN)
                            && !contentType.startsWith(MimeTypeConstants.TEXT_X_JAVA_PROPERTIES)) {
                        throw new InvalidContentTypeException(propertiesUrl, contentType);
                    }
                } else {
                    throw new MissingContentTypeException(propertiesUrl);
                }
            }

            propertiesStream = httpUrlConnection.getInputStream();
        } else {
            propertiesStream = propertiesUrl.openStream();
        }

        try {
            final File result = File.createTempFile("jbi", ".properties");
            result.deleteOnExit();
            final OutputStream output = new FileOutputStream(result);
            try {
                IOUtils.copy(propertiesStream, output);
                return result;
            } finally {
                output.close();
            }
        } finally {
            propertiesStream.close();
        }
    }

    private static final File downloadArtifactToTempZipFile(final URL artifactUrl) throws HttpRemoteArtifactException,
            IOException {

        final InputStream artifactStream;
        if ("http".equalsIgnoreCase(artifactUrl.getProtocol())) {
            final HttpURLConnection httpUrlConnection = (HttpURLConnection) artifactUrl.openConnection();
            final int responseCode = httpUrlConnection.getResponseCode();
            if (responseCode != HttpURLConnection.HTTP_OK) {
                throw new HttpCodeNotOkException(artifactUrl, responseCode);
            } else {
                // We check that the resource is a binary file
                final String contentType = httpUrlConnection.getContentType();
                if (contentType != null) {
                    if (!contentType.startsWith(MimeTypeConstants.APPLICATION_X_ZIP)
                            && !contentType.startsWith(MimeTypeConstants.APPLICATION_X_ZIP_COMPRESSED)
                            && !contentType.startsWith(MimeTypeConstants.APPLICATION_ZIP)
                            && !contentType.startsWith(MimeTypeConstants.APPLICATION_ZIP_COMPRESSED)) {
                        throw new InvalidContentTypeException(artifactUrl, contentType);
                    }
                } else {
                    throw new MissingContentTypeException(artifactUrl);
                }
            }

            artifactStream = httpUrlConnection.getInputStream();
        } else {
            artifactStream = artifactUrl.openStream();
        }

        try {
            final File result = File.createTempFile("jbi", ".zip");
            result.deleteOnExit();
            final OutputStream output = new FileOutputStream(result);
            try {
                IOUtils.copy(artifactStream, output);
                return result;
            } finally {
                output.close();
            }
        } finally {
            artifactStream.close();
        }
    }

    /**
     * Get the JBI name of this artifact
     * 
     * @return the JBI name (not null)
     * @throws ArtifactNameMissingInJBIDescriptorException
     *             The artifact name is missing in the JBI descriptor of the JBI archive
     */
    private static final String getName(final Node jbiConfig, final URL zipArchiveUrl)
            throws ArtifactNameMissingInJBIDescriptorException {
        try {
            XPath xp = XPathFactory.newInstance().newXPath();
            String result = (String) xp.evaluate(XPATH_JBI_NAME, jbiConfig, XPathConstants.STRING);
            if (result == null || result.isEmpty()) {
                throw new ArtifactNameMissingInJBIDescriptorException(zipArchiveUrl);
            }
            return result;
        } catch (XPathExpressionException xpee) {
            throw new UncheckedException(xpee);
        }
    }

    /**
     * Get the version of a shared library
     * 
     * @return the version of the shared library (not null)
     * @throws ArtifactVersionMissingInJBIDescriptorException
     *             The artifact version is missing in the JBI descriptor of the JBI archive
     */
    private static final String getSlVersion(final Node jbiConfig, final URL zipArchiveUrl)
            throws ArtifactSlVersionMissingInJBIDescriptorException {
        try {
            final XPath xp = XPathFactory.newInstance().newXPath();
            final String result = (String) xp.evaluate(XPATH_JBI_SL_VERSION, jbiConfig, XPathConstants.STRING);
            if (result == null || result.isEmpty()) {
                throw new ArtifactSlVersionMissingInJBIDescriptorException(zipArchiveUrl);
            }
            return result;
        } catch (XPathExpressionException xpee) {
            throw new UncheckedException(xpee);
        }
    }

}
