/**
 * Copyright (c) 2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.admin.api;

import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.IOException;
import java.net.URL;
import java.util.Enumeration;

import org.junit.Test;
import org.ow2.petals.admin.api.conf.PetalsAdminConfiguration;
import org.ow2.petals.admin.api.exception.DuplicatedServiceException;
import org.ow2.petals.admin.api.exception.MissingServiceException;

/**
 * 
 * @author "Mathieu CARROLLE - EBM WebSourcing"
 * 
 */
public class PetalsAdministrationFactoryTest {

    /**
     * Dummy implementation of {@link PetalsAdministrationFactory} for
     * duplication test.
     * 
     * @author Christophe DENEUX - Linagora
     * 
     */
    public class PetalsAdministrationFactoryDummy extends PetalsAdministrationFactory {

        @Override
        public PetalsAdministration newPetalsAdministrationAPI() {
            // TODO Auto-generated method stub
            return null;
        }

        @Override
        public PetalsAdministration newPetalsAdministrationAPI(PetalsAdminConfiguration config) {
            // TODO Auto-generated method stub
            return null;
        }

    }

    /**
     * <p>
     * Check the loading of the Petals Administration API implementation when only one implementation is available on
     * the classpath.
     * </p>
     * <p>
     * Expected result: The implementation available on the classpath is loaded.
     * </p>
     */
    @Test
    public void testNewInstance_00() throws DuplicatedServiceException,
            MissingServiceException {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(new ClassLoader(oldClassloader) {

            @Override
            public Enumeration<URL> getResources(final String name) throws IOException {
                if (name.equals("META-INF/services/" + PetalsAdministrationFactory.class.getName())) {
                    return super.getResources("services/mock-impl/"
                            + PetalsAdministrationFactory.class.getName());
                } else {
                    return super.getResources(name);
                }
            }
        });

        try {
            final PetalsAdministrationFactory petalsAdministrationFactory = PetalsAdministrationFactory.getInstance();
            assertTrue(petalsAdministrationFactory instanceof PetalsAdministrationFactoryMock);
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
            PetalsAdministrationFactory.close();
        }
    }

    /**
     * <p>
     * Check the loading of the Petals Administration API implementation when no
     * implementation is available on the classpath and the property is not set.
     * </p>
     * <p>
     * Expected result: The exception {@link MissingServiceException} is thrown.
     * </p>
     */
    @Test(expected = MissingServiceException.class)
    public void testNewInstance_01() throws DuplicatedServiceException,
            MissingServiceException {

        try {
            PetalsAdministrationFactory.getInstance();
            fail("Exception " + MissingServiceException.class.getName() + " is not thrown !");
        } finally {
            PetalsAdministrationFactory.close();
        }
    }

    /**
     * <p>
     * Check the loading of the Petals Administration API implementation when no
     * implementation is available on the classpath and the property is set.
     * </p>
     * <p>
     * Expected result: The implementation defined by the property is loaded.
     * </p>
     */
    @Test
    public void testNewInstance_02() throws DuplicatedServiceException,
            MissingServiceException {

        System.setProperty(PetalsAdministrationFactory.PROPERTY_NAME,
                PetalsAdministrationFactoryMock.class.getName());
        try {
            final PetalsAdministrationFactory petalsAdministrationFactory = PetalsAdministrationFactory.getInstance();
            assertTrue(petalsAdministrationFactory instanceof PetalsAdministrationFactoryMock);
        } finally {
            System.setProperty(PetalsAdministrationFactory.PROPERTY_NAME, "");
            PetalsAdministrationFactory.close();
        }
    }

    /**
     * <p>
     * Check the loading of the Petals Administration API implementation when
     * several implementations are available on the classpath.
     * </p>
     * <p>
     * Expected result: The exception {@link DuplicatedServiceException} is
     * thrown.
     * </p>
     */
    @Test(expected = DuplicatedServiceException.class)
    public void testNewInstance_03() throws DuplicatedServiceException,
            MissingServiceException {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(new ClassLoader(oldClassloader) {

            @Override
            public Enumeration<URL> getResources(final String name) throws IOException {
                if (name.equals("META-INF/services/" + PetalsAdministrationFactory.class.getName())) {
                    return super.getResources("services/duplicated/"
                            + PetalsAdministrationFactory.class.getName());
                } else {
                    return super.getResources(name);
                }
            }
        });

        try {
            PetalsAdministrationFactory.getInstance();
            fail("Exception " + DuplicatedServiceException.class.getName() + " is not thrown !");
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
            PetalsAdministrationFactory.close();
        }
    }

    /**
     * <p>
     * Check the loading of the Petals Administration API implementation when
     * only one implementation is defined into the classpath, but the class does
     * not exist.
     * </p>
     * <p>
     * Expected result: The exception {@link MissingServiceException} is thrown.
     * </p>
     */
    @Test(expected = MissingServiceException.class)
    public void testNewInstance_04() throws DuplicatedServiceException, MissingServiceException {

        final ClassLoader oldClassloader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(new ClassLoader(oldClassloader) {

            @Override
            public Enumeration<URL> getResources(final String name) throws IOException {
                if (name.equals("META-INF/services/" + PetalsAdministrationFactory.class.getName())) {
                    return super.getResources("services/class-doesnot-exist/"
                            + PetalsAdministrationFactory.class.getName());
                } else {
                    return super.getResources(name);
                }
            }
        });

        try {
            PetalsAdministrationFactory.getInstance();
            fail("Exception " + MissingServiceException.class.getName() + " is not thrown !");
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassloader);
            PetalsAdministrationFactory.close();
        }
    }
}
